//  **** BEGIN LICENSE BLOCK ****
//  Copyright(c) 2005 Adam Judson
//
//  PhotoMap
//  - display user contributed photos on maps.
//
//  This program is free software; you can redistribute it and/or modify it under
//  the terms of the GNU General Public License as published by the Free
//  Software Foundation; either version 2 of the License, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
//  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
//  more details.
//
//  You should have received a copy of the GNU General Public License along with
//  this program; if not, write to the Free Software Foundation, Inc., 59 Temple
//  Place, Suite 330, Boston, MA 02111-1307 USA
//
//  **** END LICENSE BLOCK ****


// these are the classes needed to run the google earth server

function PhotoMapServer(controller) {
   this.controller = controller;
   this.preferences = controller.preferences;
}

PhotoMapServer.start = function() {
   var controller = window.arguments[0];
   PhotoMapServer.instance = new PhotoMapServer(controller);
   PhotoMapServer.instance.start();
}

PhotoMapServer.stop = function() {
   PhotoMapServer.instance.stop();
}

PhotoMapServer.prototype = {
   start : function() {
      if (this.serverSocket) {
         return;
      }

      try {
         this.serverSocket = Components.classes["@mozilla.org/network/server-socket;1"].createInstance(Components.interfaces.nsIServerSocket);
         this.serverSocket.init(this.preferences.getEarthServerPort(), true,-1);
         this.serverSocket.asyncListen(this);
         document.getElementById("photomapserver.stop").removeAttribute("disabled");
         document.getElementById("photomapserver.start").setAttribute("disabled", true);
      } catch(ex) {
         window.alert(ex);
      }
  },

   stop : function() {
      if(this.serverSocket) {
         try {
            this.serverSocket.close();
         } catch(ex) {
         }
         this.serverSocket = null;
         this.serverRunning = false;
      }
   },
  
   restart : function() {
      this.stopServer();
      this.startServer();
   },

   clear : function() {
      document.getElementById("photomapserver.log").value = "";
   },

   launch : function() {
      var url = "http://127.0.0.1:" + this.preferences.getEarthServerPort() + "/map.kml?" + (new Date()).getTime();
      var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService(Components.interfaces.nsIWindowMediator);
      var win = wm.getMostRecentWindow("navigator:browser");
      win.gBrowser.selectedTab = win.gBrowser.addTab(url);
   },

   // nsIServerSocketListener functions
   onSocketAccepted : function onSocketAccepted(serverSocket, transport) {
      var request = new PhotoMapServerRequest(this.controller, transport);
      request.process();
   },

   onStopListening : function onStopListening(serverSocket, status) {
      if (document) {
         var log = document.getElementById("photomapserver.log");
         if (log) {  
            log.value = log.value + "Stopped\r\n";;
         }
         document.getElementById("photomapserver.start").removeAttribute("disabled");
         document.getElementById("photomapserver.stop").setAttribute("disabled", true);
      }
      this.stop();
   },
  // end nsIServerSocketListener

   QueryInterface: function (iid) {
      if (iid.equals(Components.interfaces.nsISupports)) {
         return this;
      }
      throw Components.results.NS_ERROR_NO_INTERFACE;
   }
};

function PhotoMapServerRequest(controller, transport) {
   this.controller = controller;
   this.agent = controller.photo;
   this.preferences = this.agent.preferences;
   this.transport = transport;
   this.outputWindow = document.getElementById("photomapserver.log");
}

PhotoMapServerRequest.prototype = {
   process : function() {
      this.iStream = this.transport.openInputStream(0,0,0);
      var scriptableDataStream = PhotoMapUtils.createScriptableInputStream(this.iStream);

      var dataListener = new PhotoMapServerDataListener(scriptableDataStream, this);

      var pump = Components.classes["@mozilla.org/network/input-stream-pump;1"].createInstance(Components.interfaces.nsIInputStreamPump);
      pump.init(this.iStream, -1, -1, 0, 0, true);
      pump.asyncRead(dataListener, null);

   },

   finished : function() {
      if (this.iStream) {
         this.iStream.close();
      }
      if (this.oStream) {
         this.oStream.close();
      }
      this.logOutput("Closed");
   },

   handleData: function(data) {
      // this.logInput(data);
      this.oStream = this.transport.openOutputStream(0,0,0);
      if (this.isEmptyRequest(data)) {
         this.logInput("Empty Request");
         this.writeNetworkLink();
      } else {
         var inputBBOX = this.parseInput(data);
         this.logInput(inputBBOX.toSource());
         var photos = this.getResponseData(inputBBOX);
         
         var response = "HTTP/1.1 200 OK\r\n";
         response += "Content-Type: application/vnd.google-earth.kml+xml\r\n"; 
         response += "\r\n"; 
         response += '<?xml version="1.0" encoding="UTF-8"?><kml xmlns="http://earth.google.com/kml/2.0">';
         response += this.buildResponseElements(photos);
         response += "</kml>\r\n\r\n";
         this.oStream.write(response, response.length);
      }
      this.oStream.flush();
      this.logOutput("..");
      // this.logOutput(response);
      this.transport.close(0);
   },

   isEmptyRequest : function(data) {
      return (data.indexOf("BBOX") == -1);
   },

   writeNetworkLink : function() {
      var header = "HTTP/1.1 200 OK\r\n" +
                   "Content-Type: application/vnd.google-earth.kml+xml\r\n" +
                   "\r\n"; 

      var networkLinkKML = '<?xml version="1.0" encoding="UTF-8"?>\n' +
                           '<kml xmlns="http://earth.google.com/kml/2.0">\n' +
                           '<Folder>\n<name>PhotoMap places</name>' +
                           '   <NetworkLink>\n' +
                           '      <name>PhotoMap network link</name>\n' +
                           '      <description>Show images in flickr saved with <a href="http://photomap.mozdev.org">PhotoMap</a>.  Make sure you have launched the server from Firefox. (Tools->PhotoMap->Google Earth)</description>\n' +
                           '      <Url>\n' +
                           '         <href>http://127.0.0.1:' + this.preferences.getEarthServerPort() + '/map.kml</href>\n' +
                           '         <viewRefreshMode>onStop</viewRefreshMode>\n' +
                           '         <viewRefreshTime>1</viewRefreshTime>\n' + 
                           '      </Url>\n';
      if (this.preferences.getServerForceDisplay()) {
         networkLinkKML += '      <refreshVisibility>1</refreshVisibility>\n';
      }

      networkLinkKML += '   </NetworkLink>\n';

      // move to our current position
      var loc = this.controller.getLocation();
      networkLinkKML += '   <LookAt>\n' +
                        '      <longitude>' + loc.lon + '</longitude>\n' +
                        '      <latitude>'  + loc.lat + '</latitude>\n'  +
                        '      <range>' + (Math.max(1, loc.zoom) * 250.0) + '</range>\n'  +
                        '      <tilt>0</tilt>\n' +
                        '      <heading>0</heading>\n' +
                        '   </LookAt>\n';
      networkLinkKML += '</Folder>\n</kml>';   
      var response = header + networkLinkKML + "\r\n\r\n";
      this.oStream.write(response, response.length);
   },

   buildResponseElements : function(photos) {
      var response = "<Folder><name>PhotoMap</name><description>Photos from PhotoMap</description><visibility>1</visibility>";
      var photoIndex, photo;
      var firstPhoto = null;
      var pov;
      var sortedPhotos;
      for (var p = 0; p < photos.length; p++) {
         sortedPhotos = Photo.sort(photos[p]);
         // check if photos is empty?
         response += "<Placemark>\n";
         // now pull out the first photo.
         for (photoIndex in sortedPhotos) {
            photo = sortedPhotos[photoIndex];
            if (!firstPhoto) {
               firstPhoto = photo;
               response += "<Snippet>Photo From " + photo.provider + ": " + photo.lat + ", " + photo.lon + "</Snippet>";
               response += "<description><![CDATA[";
            }
            pov = (photo.pov == "unknown") ? "who knows where" : photo.pov;
            response += "<p><h4>Facing " + pov + "</h4><a href='" + photo.url + "'>";
            response += "<img src='" + photo.mediumUrl + " ' border='0'></img></a>";
         }
         response += "]]></description>\n";
         // names are not really helpful
         // response += "<name>" + firstPhoto.title + "</name>";
         response += "<name></name>";
         response += "<Style>\n<IconStyle>\n<Icon>\n<href>" + firstPhoto.smallUrl + "</href>\n";
         // FIXME these are not the right sizes...
         response += "<x>0</x>\n<y>0</y>\n<w>75</w>\n<h>75</h>\n</Icon>\n</IconStyle>\n</Style>";
         response += "<Point>\n<extrude>1</extrude><coordinates>" + firstPhoto.lon + "," + firstPhoto.lat + ",1</coordinates>\n";
         response += "</Point>\n</Placemark>\n";
         firstPhoto = null;
         
      }
      response += "</Folder>\n";
      return response;
   },

   parseInput : function(data) {
      var start = data.indexOf("BBOX=") + 5;
      data = data.substring(start);
      var end = data.indexOf(" ");
      var bbox = data.substring(0, end);
      var boxVals = bbox.split(",");
      return {minX : boxVals[0], minY : boxVals[1], maxX : boxVals[2], maxY : boxVals[3]};
   },

   getResponseData : function(bounds) {
      var zoom = this.calculateZoom(bounds);
      var loc  = this.calculateLocation(bounds);
      this.agent.locationChanged(loc, bounds, zoom, false);
      /*
      var photoKeys = this.agent.getCurrentResults(loc, bounds, zoom);
      var currentPhotos;
      for (var i = 0; i < photoKeys.length; i++) {
         currentPhotos = photoKeys[i];
      }
      */
      // this is not very scalable...
      var photos = new Array();
      var photoKeys = this.agent.getAll();
      var photo;
      for (var key in photoKeys) {
         // we are really only looking at the first one...
         for (var photo in photoKeys[key]) {
            if (this.inBounds(photoKeys[key][photo], bounds, zoom)) {
               photos.push(photoKeys[key]);
            }
            break;
         }
      }
      return photos;
   },

   inBounds : function(photo, bounds, zoom) {
      if (photo.type >= zoom) {
         if (parseFloat(photo.lat) < bounds.maxY && 
             parseFloat(photo.lat) > bounds.minY && 
             parseFloat(photo.lon) < bounds.maxX &&
             parseFloat(photo.lon) > bounds.minX) {
            this.log("Photo in bounds - " + photo.title);
            return true;
         }
      }
      return false;
      // return true;
   },

   calculateZoom : function(bounds) {
      var zoom = 9;
      var delta = Math.min(Math.abs(bounds.maxX - bounds.minX), Math.abs(bounds.maxY - bounds.minY));
      if (delta > 20) {
         zoom = 9;
      } else if (delta > 7) {
         zoom = 8;
      } else if (delta > 2) {
         zoom = 7;
      } else if (delta > 0.5) {
         zoom = 6;
      } else if (delta > 0.1) {
         zoom = 5;
      } else if (delta > 0.03) {
         zoom = 4;
      } else if (delta > 0.0075) {
         zoom = 3;
      } else if (delta > 0.003) {
         zoom = 2;
      } else if (delta <= 0.003) {
         zoom = 1;
      }
      this.log("Zoom is: " + zoom + ", delta is: " + delta);
      return zoom;
   },

   calculateLocation : function(bounds) {
      var lat = (parseFloat(bounds.minY) + parseFloat(bounds.maxY))/2;
      var lon = (parseFloat(bounds.minX) + parseFloat(bounds.maxX))/2;
      return {lat : lat, lon : lon};
   },

   log : function(data) {
      this.outputWindow.value = this.outputWindow.value + "\r\n" + data + "\r\n";
   },

   logInput : function(data) {
      if (this.outputWindow.value > 5000) {
         this.outputWindow.value = "";
      }
      this.outputWindow.value = this.outputWindow.value + "------------\r\n" + data + "\r\n";
   },

   logOutput : function(data) {
      this.outputWindow.value = this.outputWindow.value + "=================================\r\n" + data + "\r\n";
   }
};

function PhotoMapServerDataListener(stream, request) {
   this.stream = stream;
   this.request = request;
}

PhotoMapServerDataListener.prototype = {
   data : "",

   onStartRequest : function(request, context) {
      this.data = "";
   },

   onStopRequest : function(request, context, status){
      this.stream = null;
      this.request.finished();
   },

   onDataAvailable : function(request, context, inputStream, offset, count) {
      try {
         this.data += this.stream.read(count);
         if (this.data.length >= 4 && (this.data.indexOf("\r\n\r\n") == (this.data.length - 4))) {
            this.request.handleData(this.data);
            this.data = "";
         }
      } catch(e) {
         PhotoMapUtils.log(e);
      }
   },

   QueryInterface : function (iid) {
      if (iid.equals(Components.interfaces.nsISupports) ||
          iid.equals(Components.inferfaces.nsIStreamListener) ||
          iid.equals(Components.interfaces.nsIRequestObserver)) {
         return this;
      }
      throw Components.results.NS_ERROR_NO_INTERFACE;
   }
};
